"""elementary.py
This file overloads elementary functions which do no have dunder methods
including exponential and trig functions. The functions first try to work 
with x as a dual number and falls back to treating x as a normal numerical type"""

from pydif.dual.dual import Dual 
import numpy as np 

def cos(x):
    try: 
        return Dual(np.cos(x.val), -1 * x.der * np.sin(x.val), x.der**2 * -1 * np.cos(x.val)- x.der2 * np.sin(x.val))
    except:
        return np.cos(x)

def sin(x):
    try: 
        return Dual(np.sin(x.val),  x.der * np.cos(x.val), x.der2 * np.cos(x.val) - x.der **2 * np.sin(x.val))
    except:
        return np.sin(x)

def tan(x):
    try:
        return Dual(np.tan(x.val), x.der * (1/np.cos(x.val))**2, (1/np.cos(x.val)**2) * (x.der2 + 2*x.der**2 * np.tan(x.val)))
    except:
        return np.tan(x)

def arccos(x):
    try:
        return Dual(np.arccos(x.val), -1 * x.der * 1/(np.sqrt(1- x.val**2)), -1 * (x.val**2*-1*(x.der2) + x.der2 + x.val * x.der**2)/(1- x.val**2)**(3/2))
    except:
        return np.arccos(x)


def arcsin(x):
    try:
        return Dual(np.arcsin(x.val), x.der * 1/(np.sqrt(1- x.val**2)), (x.val*x.der**2 - (x.val**2 -1 )*x.der2)/(1-x.val**2)**(3/2))
    except:
        return np.arcsin(x)


def arctan(x):
    try:
        return Dual(np.arctan(x.val), x.der * 1/(x.val**2 +1), ((x.val**2+1)*x.der2 - 2 * x.val * x.der**2)/(x.val**2 +1)**2)
    except:
        return np.arctan(x)

def sinh(x):
    try:
        return Dual(np.sinh(x.val), x.der * np.cosh(x.val), x.der2 * np.cosh(x.val) + x.der **2 * np.sinh(x.val))
    except:
        return np.sinh(x)

def cosh(x):
    try:
        return Dual(np.cosh(x.val), x.der * np.sinh(x.val), x.der2 * np.sinh(x.val) + x.der **2 *np.cosh(x.val))
    except:
        return np.cosh(x)

def tanh(x):
    try:
        return Dual(np.tanh(x.val), x.der * 1/(np.cosh(x.val)**2), (1/np.cosh(x.val))**2 * (x.der2 - 2*x.der**2 * np.tanh(x.val)))
    except:
        return np.tanh(x)

def arcsinh(x):
    try:
        return Dual(np.arcsinh(x.val), x.der * 1/np.sqrt(x.val**2 +1), (x.val**2 * x.der2 + x.der2 - x.val*x.der**2)/(x.val**2 + 1)**(3/2))
    except:
        return np.arcsinh(x)

def arccosh(x):
    try:
        return Dual(np.arccosh(x.val), x.der * (1/np.sqrt(x.val -1) * 1/(np.sqrt(x.val +1))), (x.val**2*x.der2 - x.der2 - x.val*x.der**2)/((x.val-1)**(3/2) *(x.val+1)**(3/2)))
    except:
        return np.arccosh(x)

def arctanh(x):
    try:
        return Dual(np.arctanh(x.val), -1 * x.der * 1/(x.val**2 -1), (-1*x.val**2*x.der2 + x.der2+ 2 * x.val * x.der**2)/(1-x.val**2)**2)
    except:
        return np.arctanh(x)

def exp(x):
    try:
        return Dual(np.exp(x.val), x.der * np.exp(x.val), np.exp(x.val) * (x.der2 + x.der**2))
    except:
        return np.exp(x)

def exp2(x):
    try:
        return Dual(np.exp2(x.val), np.exp2(x.val) * (x.der * np.log(2)), np.log(2)*np.exp2(x.val)*(x.der2 + np.log(2) * x.der**2))
    except:
        return np.exp2(x)

# natural log 
def log(x):
    try: 
        return Dual(np.log(x.val), (1 / float(x.val)) * x.der, (x.val * x.der2 - x.der**2)/(x.val**2)  )
    except:
        return np.log(x)    

# log base 2
def log2(x):
    try:
        return Dual(np.log2(x.val), 1/(x.val * np.log(2)) * x.der, (x.val * x.der2- x.der**2)/(np.log(2)* x.val**2))
    except:
        return np.log2(x)

# log base 10
def log10(x):
    try:
        return Dual(np.log10(x.val), 1/(x.val * np.log(10)) * x.der, (x.val * x.der2 - x.der**2)/(np.log(10)*x.val**2))
    except:
        return np.log10(x)


def sqrt(x):
    try:
        return Dual(np.sqrt(x.val), x.der * 1/(2*np.sqrt(x.val)), (2 * x.val * x.der2 - x.der**2)/(4 * x.val **(3/2)))
    except:
        return np.sqrt(x)

def sigmoid(x):
    try:
        a = x.val
        return 1/(1 + exp(-1*x))
    except:
        return 1/(1+ np.exp(-1 * x))

def relu(x):
    try:
        u = np.maximum(0, x.val)
        v = np.where(u > 0, 1, 0)
        return Dual(u, x.der * v, 0)
    except:
        return np.maximum(0, x)




"""pydif.py
This file implements the autodiff class. The autodiff class is initialized with a function that accepts single variables of the form f(x,y,z,...).
The autodiff object then allows for the evaluation of the value, and derivatives of a function at a specified position.

Args:
    param1 (function object): The function whose value and derivatives are to be evaluated.

Returns:
    autodiff Object: Returns a new autodiff object for evaluation of the specified function.

"""

from inspect import signature
from pydif.dual.dual import Dual
import numpy as np
import collections

class autodiff():

    #initialize autodiff class with a function
    def __init__(self, func):
        self.func = func
        self.num_params = len(signature(func).parameters)

    #make function parameters dual numbers and evaluate function at a position
    def _eval(self, pos, wrt_variables = False):
        if isinstance(pos, collections.Iterable):
            if wrt_variables:
                params = []
                for cursor, pos_i in enumerate(pos):
                    der_partials = np.zeros(self.num_params)
                    der_partials[cursor] = 1
                    params.append(Dual(pos_i, der_partials, np.zeros(self.num_params)))
            else:
                params = [Dual(pos_i,1,0) for pos_i in pos]
            return self.func(*params)
        else:
            if wrt_variables:
                params = Dual(pos,[1],[0])
            else:
                params = Dual(pos,1,0)
            return self.func(params)

    #check that the specified iterable is the same shape as the function (specified at object creation) input
    def _check_dim(self, item):
        badDimentionsMsg = 'poorly formatted position or direction. should be of length {}.'.format(self.num_params)
        if isinstance(item, collections.Iterable):
            if len(item) != self.num_params:
                raise ValueError(badDimentionsMsg)
        else:
            if self.num_params != 1:
                raise ValueError(badDimentionsMsg)

    def _enforce_unitvector(self, direction):
        tmp_direction = np.array(direction)
        magnitude = np.sqrt(tmp_direction.dot(tmp_direction))
        if magnitude == 0:
            raise ValueError('poorly formatted direction. should be a vector of non-zero magnitude.')
        if magnitude != 1:
            tmp_direction = tmp_direction / magnitude
        return tmp_direction

    #evaluate the value of the function at a specified position
    def get_val(self, pos, direction=None):
        if direction is None:
            self._check_dim(pos)
            res = self._eval(pos)
            if isinstance(res, collections.Iterable):
                return [i.val for i in res]
            else:
                return res.val
        else:
            self._check_dim(pos)
            self._check_dim(direction)
            self._enforce_unitvector(direction)
            res = self._eval(pos)
            if isinstance(res, collections.Iterable):
                return np.sum(np.array([i.val for i in res]) * direction)
            else:
                return res.val

    #evaluate the derivatives of the function at a specified position in a specified direction
    def get_der(self, pos, wrt_variables = False, direction=None, order=1):
        if order == 1:
            der_attr = 'der'
        elif order == 2:
            der_attr = 'der2'
        else:
            raise NotImplementedError('only first and second order derivatives are implemented.')

        if direction is None:
            self._check_dim(pos)
            res = self._eval(pos, wrt_variables)

            if isinstance(res, collections.Iterable):
                return [getattr(i, der_attr) for i in res]
            else:
                return getattr(res, der_attr)
        else:
            self._check_dim(pos)
            self._check_dim(direction)
            self._enforce_unitvector(direction)
            res = self._eval(pos, wrt_variables)

            if isinstance(res, collections.Iterable):
                return np.sum(np.array([getattr(i, der_attr) for i in res]) * direction)
            else:
                return getattr(res, der_attr)

class autodiff_vector():

    #initialize autodiff vector with autodiff objects
    def __init__(self, funcs):
        funcs = np.array(funcs)
        self.func_vector = np.empty(funcs.shape, dtype=object)
        for cursor, func in np.ndenumerate(funcs):
            self.func_vector[cursor] = autodiff(func)

    def _clean_dim(self, param, req_dim):
        if isinstance(param, collections.Iterable):
            return param[:req_dim]
        else:
            return param

    #evaluate the value of the vector function at a specified position
    def get_val(self, pos, direction=None):
        res = np.empty(self.func_vector.shape, dtype=object)
        for cursor, autodiff_obj in np.ndenumerate(self.func_vector):
            clean_pos = self._clean_dim(pos, self.func_vector[cursor].num_params)
            clean_direction = self._clean_dim(direction, self.func_vector[cursor].num_params)
            res[cursor] = autodiff_obj.get_val(clean_pos, clean_direction)
        return res

    #evaluate the derivative of the vector function at a specified position in a specified direction
    def get_der(self, pos, wrt_variables = False, direction=None, order=1):
        res = np.empty(self.func_vector.shape, dtype=object)
        for cursor, autodiff_obj in np.ndenumerate(self.func_vector):
            clean_pos = self._clean_dim(pos, self.func_vector[cursor].num_params)
            clean_direction = self._clean_dim(direction, self.func_vector[cursor].num_params)
            res[cursor] = autodiff_obj.get_der(clean_pos, wrt_variables, clean_direction, order)
        return res

def _eval_func(func, args, der_attr):
    if isinstance(args, collections.Iterable):
        params = []
        for cursor, arg in enumerate(args):
            der_partials = np.zeros(len(args))
            der_partials[cursor] = 1
            params.append(Dual(arg, der_partials, np.zeros(len(args))))
        res = func(*params)
        if isinstance(res, collections.Iterable):
            return [getattr(i, der_attr) for i in res]
        else:
            return getattr(res, der_attr)
    else:
        args = Dual(args,[1],[0])
        res = func(args)
        if isinstance(res, collections.Iterable):
            return [getattr(i, der_attr) for i in res]
        else:
            return getattr(res, der_attr)

def diff(func):
    def func_wrapper(args):
        return _eval_func(func, args, 'der')
    return func_wrapper

def diffdiff(func):
    def func_wrapper(args):
        return _eval_func(func, args, 'der2')
    return func_wrapper

if __name__ == '__main__':
    pass



import pytest
import numpy as np
from pydif.optimize.optimize import Optimize

# global minimum at (1, 1)
def rosenbrock(x, y):
    return ((100)*((y - (x**2))**2)) + ((1-x)**2)

# global minimum at (3, 2)
def f1(x, y):
    return x**2 - 6 * x + y**2 - 4 * y + 2

# global minimum at (0, 0)
def f2(x, y):
    return  x**2 + 4 * (y**2) - 2 * (x**2) * y + 4

# global minimum at (0, 0)
def f3(x, y):
    return  x**2 + x * y + y**2

# roots at (0, 1) amd (0, -1)
def f4(x, y):
    return np.array(
        [((x**4 + y**4)**(1/4)) - 1,
         ((((3*x) - (y))**4 + (x)**4)**(1/4)) - 1])

def test_gradient_descent():

    min_pos = Optimize(f1).gradient_descent((1, 1))
    assert(min_pos[0] == pytest.approx(3, rel = 1e-2, abs = 1e-2))
    assert(min_pos[1] == pytest.approx(2, rel = 1e-2, abs = 1e-2))

    min_pos = Optimize(f2).gradient_descent((1, 1))
    assert(min_pos[0] == pytest.approx(0, rel = 1e-2, abs = 1e-2))
    assert(min_pos[1] == pytest.approx(0, rel = 1e-2, abs = 1e-2))

    min_pos = Optimize(f3).gradient_descent((1, 1))
    assert(min_pos[0] == pytest.approx(0, rel = 1e-2, abs = 1e-2))
    assert(min_pos[1] == pytest.approx(0, rel = 1e-2, abs = 1e-2))

def test_BFGS():

    min_pos = Optimize(f1).BFGS((1, -1))
    assert(min_pos[0] == pytest.approx(3, rel = 1e-2, abs = 1e-2))
    assert(min_pos[1] == pytest.approx(2, rel = 1e-2, abs = 1e-2))

    min_pos = Optimize(f2).BFGS((1, -1))
    assert(min_pos[0] == pytest.approx(0, rel = 1e-2, abs = 1e-2))
    assert(min_pos[1] == pytest.approx(0, rel = 1e-2, abs = 1e-2))

    min_pos = Optimize(f3).BFGS((1, -1))
    assert(min_pos[0] == pytest.approx(0, rel = 1e-2, abs = 1e-2))
    assert(min_pos[1] == pytest.approx(0, rel = 1e-2, abs = 1e-2))

def test_newton():

    root = Optimize(f4).newton(np.array([-0.5, 1.5]), max_iters = 10000)
    assert(root[0] == pytest.approx(0, rel = 1e-2, abs = 1e-2))
    assert(root[1] == pytest.approx(1, rel = 1e-2, abs = 1e-2))

    root = Optimize(f4).newton(np.array([0.5, -1.5]), max_iters = 10000)
    assert(root[0] == pytest.approx(0, rel = 1e-2, abs = 1e-2))
    assert(root[1] == pytest.approx(-1, rel = 1e-2, abs = 1e-2))



import sys
import os
import numpy as np
from inspect import signature
from collections import Iterable
sys.path.append(os.path.join(os.getcwd(),'pydif'))
np.warnings.filterwarnings('ignore')
from pydif.pydif import autodiff

class Optimize():
    def __init__(self, func):
        self.func = func
        self.num_params = len(signature(func).parameters)

    def gradient_descent(self, init_pos, step_size=0.1, max_iters=100, precision=0.001, return_hist=False):
        num_params = len(signature(self.func).parameters)
        badDimentionsMsg = 'poorly formatted initial position. should be of length {}.'.format(num_params)
        if num_params != len(init_pos):
            raise ValueError(badDimentionsMsg)

        cur_pos = init_pos
        iters = 0
        dfdx = autodiff(self.func)
        val = dfdx.get_val(init_pos)

        if isinstance(val, Iterable):
            raise ValueError("The optimize class only optimizes scalar valued functions")

        # pre allocate history array
        hist = [cur_pos]
        prev_step_size = 100 + precision
        while (prev_step_size > precision and iters < max_iters):
            jac = dfdx.get_der(cur_pos, wrt_variables=True)
            prev_pos = cur_pos
            cur_pos = cur_pos - step_size * jac
            prev_step_size = np.linalg.norm(abs(cur_pos - prev_pos))
            iters += 1
            hist.append(cur_pos) #store history
        np.array(hist)

        if return_hist:
            return cur_pos, hist
        else:
            return cur_pos

    def delta_B(self, y, s, B):
        y = y.reshape((self.num_params, 1))
        s = s.reshape((self.num_params, 1))
        return np.matmul(y, y.T)/np.matmul(y.T, s)-np.matmul(np.matmul(np.matmul(B, s), s.T), B)/(np.matmul(np.matmul(s.T, B),s))

    def BFGS(self, init_pos, max_iters=100, precision=10**-8, return_hist=False):
        #set original values
        coord = init_pos

        hist = [coord] #preallocate array

        #set inital conditions
        s = 0
        iterations = 0
        step = 10000
        B = np.identity(self.num_params)
        dfdx = autodiff(self.func)

        #set conditions for while loop
        while ((iterations <= max_iters) and (step >= precision)):
            s = np.linalg.solve(B, -dfdx.get_der(coord, wrt_variables = True)) #step
            step = np.linalg.norm(s) #step size
            y = dfdx.get_der(coord+s, wrt_variables = True) - dfdx.get_der(coord, wrt_variables = True) #new y
            coord = coord + s #new coord
            B = B + self.delta_B(y, s, B) #get new B
            iterations += 1
            hist.append(coord)
        hist = np.array(hist)

        if return_hist:
            return coord, hist
        else:
            return coord

    def newton(self, init_pos, step_size=0.1, max_iters=100, precision=0.001, return_hist=False):
        num_params = len(signature(self.func).parameters)
        badDimentionsMsg = 'poorly formatted initial position. should be of length {}.'.format(num_params)
        if num_params != len(init_pos):
            raise ValueError(badDimentionsMsg)
        cur_pos = init_pos
        iters = 0
        dfdx = autodiff(self.func)
        val = dfdx.get_val(init_pos)
         # if isinstance(val, Iterable):
        #     raise ValueError("The optimize class only optimizes scalar valued functions")
         #preallocate arrays
        hist = [cur_pos]
         #set inital conditions
        iters = 0
        s_k = 0
        step = 10000
         #define conditions to break loop
        while ((iters <= max_iters) and (np.linalg.norm(step) >= precision)):
            step = np.linalg.solve(dfdx.get_der(cur_pos, wrt_variables=True, order=1), -self.func(*cur_pos))
            cur_pos = cur_pos + step
            iters += 1
            hist.append(cur_pos)
        hist = np.array(hist)
        if return_hist:
            return cur_pos, hist
        else:
            return cur_pos



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pydif",
    version="0.1.0",
    author="Hamish Nicholson, Nicholas Pagel, Manish Reddy Vuyyuru, Victor Sheng",
    author_email="hamish_nicholson@college.harvard.edu, pagel@g.harvard.edu, mvuyyuru@g.harvard.edu, victorsheng@g.harvard.edu",
    description="A work in progress package to implement automatic differentiation for numeric functions in python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pydif/cs207-FinalProject",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
